"""Module for utility functions used in other modules regarding MLST."""

__author__ = "Cetin, Oemer"

import json
from pathlib import Path
from io import StringIO
import requests
from Bio import SeqIO
from xspect.definitions import get_xspect_model_path


def create_fasta_files(locus_path: Path, fasta_batch: str) -> None:
    """
    Create Fasta-Files for every allele of a locus.

    This function creates a fasta file for each record in the batch-string of a locus.
    The batch originates from an API-GET-request to PubMLST.
    The files are named after the record ID.
    If a fasta file already exists, it will be skipped.

    Args:
        locus_path (Path): The directory where the fasta-files will be saved.
        fasta_batch (str): A string containing every record of a locus from PubMLST.
    """
    # fasta_batch = full string of a fasta file containing every allele sequence of a locus
    for record in SeqIO.parse(StringIO(fasta_batch), "fasta"):
        number = record.id.split("_")[-1]  # example id = Oxf_cpn60_263
        output_fasta_file = locus_path / f"Allele_ID_{number}.fasta"
        if output_fasta_file.exists():
            continue  # Ignore existing ones
        with open(output_fasta_file, "w", encoding="utf-8") as allele:
            SeqIO.write(record, allele, "fasta")


def pick_species_number_from_db(available_species: dict) -> str:
    """
    Get the chosen species from all available ones in the database.

    This function lists all available species of PubMLST.
    The user is then asked to pick a species by its associated number.

    Args:
        available_species (dict): A dictionary storing all available species.

    Returns:
        str: The name of the chosen species.

    Raises:
        ValueError: If the user input is not valid.
    """
    # The "database" string can look like this: pubmlst_abaumannii_seqdef
    for counter, database in available_species.items():
        print(str(counter) + ":" + database.split("_")[1])
    print("\nPick one of the above databases")
    while True:
        try:
            choice = input("Choose a species by selecting the corresponding number:")
            if int(choice) in available_species.keys():
                chosen_species = available_species.get(int(choice))
                return chosen_species
            print(
                "Wrong input! Try again with a number that is available in the list above."
            )
        except ValueError:
            print(
                "Wrong input! Try again with a number that is available in the list above."
            )


def pick_scheme_number_from_db(available_schemes: dict) -> str:
    """
    Get the chosen scheme from all available ones of a species.

    This function lists all available schemes of a species.
    The user is then asked to pick a scheme by its associated number.

    Args:
        available_schemes (dict): A dictionary storing all available schemes.

    Returns:
        str: The name of the chosen scheme.

    Raises:
        ValueError: If the user input is not valid.
    """
    # List all available schemes of a species database
    for counter, scheme in available_schemes.items():
        print(str(counter) + ":" + scheme[0])
    print("\nPick any available scheme that is listed for download")
    while True:
        try:
            choice = input("Choose a scheme by selecting the corresponding number:")
            if int(choice) in available_schemes.keys():
                chosen_scheme = available_schemes.get(int(choice))[1]
                return chosen_scheme
            print(
                "Wrong input! Try again with a number that is available in the above list."
            )
        except ValueError:
            print(
                "Wrong input! Try again with a number that is available in the above list."
            )


def scheme_list_to_dict(scheme_list: list[str]):
    """
    Converts the scheme list into a dictionary.

    Args:
        scheme_list (list[str]): A list storing all chosen schemes.

    Returns:
        dict: The converted dictionary.
    """
    return dict(zip(range(1, len(scheme_list) + 1), scheme_list))


def pick_scheme_from_models_dir() -> Path:
    """
    Get the chosen scheme from models that have been fitted prior.

    This function creates a dictionary containing all trained models.
    The dictionary is used as an argument for the "pick_scheme" function.

    Returns:
        Path: The path to the chosen model (trained).
    """
    schemes = {}
    counter = 1
    for entry in sorted((get_xspect_model_path() / "MLST").iterdir()):
        schemes[counter] = entry
        counter += 1
    return pick_scheme(schemes)


def pick_scheme(available_schemes: dict) -> Path:
    """
    Get the chosen scheme from the scheme dictionary.

    This function lists all available schemes of a species that have been downloaded.
    The user is then asked to pick a scheme by its associated number.

    Args:
        available_schemes (dict): A dictionary storing all available schemes.

    Returns:
        Path: The path to the chosen model (trained).

    Raises:
        ValueError: If the user input is not valid or if no scheme was downloaded prior.
    """
    if not available_schemes:
        raise ValueError("No scheme has been chosen for download yet!")

    if len(available_schemes.items()) == 1:
        return next(iter(available_schemes.values()))

    # List available schemes
    for counter, scheme in available_schemes.items():
        # For Strain Typing with an API-POST Request to the db
        if str(scheme).startswith("http"):
            scheme_json = requests.get(scheme, timeout=10).json()
            print(str(counter) + ":" + scheme_json["description"])

        # To pick a scheme after download for fitting
        else:
            print(str(counter) + ":" + str(scheme).rsplit("/", maxsplit=1)[-1])

    print("\nPick a scheme for strain type prediction")
    while True:
        try:
            choice = input("Choose a scheme by selecting the corresponding number:")
            if int(choice) in available_schemes.keys():
                chosen_scheme = available_schemes.get(int(choice))
                return chosen_scheme
            print(
                "Wrong input! Try again with a number that is available in the above list."
            )
        except ValueError:
            print(
                "Wrong input! Try again with a number that is available in the above list."
            )


class MlstResult:
    """Class for storing MLST results."""

    def __init__(
        self,
        scheme_model: str,
        steps: int,
        hits: dict[str, list[dict]],
        input_source: str = None,
    ):
        """Initialise an MlstResult object."""
        self.scheme_model = scheme_model
        self.steps = steps
        self.hits = hits
        self.input_source = input_source

    def get_results(self) -> dict:
        """
        Stores the result of a prediction in a dictionary.

        Returns:
            dict: The result dictionary with s sequence ID as key and the Strain type as value.
        """
        return dict(self.hits.items())

    def to_dict(self) -> dict:
        """
        Converts all attributes into one dictionary.

        Returns:
            dict: The dictionary containing all metadata of a run.
        """
        result = {
            "Scheme": self.scheme_model,
            "Steps": self.steps,
            "Results": self.get_results(),
            "Input_source": self.input_source,
        }
        return result

    def save(self, output_path: Path | str) -> None:
        """
        Saves the result as a JSON file.

        Args:
            output_path (Path,str): The path where the results are saved.
        """

        if isinstance(output_path, str):
            output_path = Path(output_path)

        output_path.parent.mkdir(exist_ok=True, parents=True)
        json_object = json.dumps(self.to_dict(), indent=4)

        with open(output_path, "w", encoding="utf-8") as file:
            file.write(json_object)



"""Module for connecting with the PubMLST database via API requests and downloading allele files."""

__author__ = "Cetin, Oemer"

import json
import requests
from xspect.mlst_feature.mlst_helper import (
    create_fasta_files,
    pick_species_number_from_db,
    pick_scheme_number_from_db,
    pick_scheme,
    scheme_list_to_dict,
)
from xspect.definitions import get_xspect_mlst_path, get_xspect_upload_path


class PubMLSTHandler:
    """Class for communicating with PubMLST and downloading alleles (FASTA-Format) from all loci."""

    base_url = "https://rest.pubmlst.org/db"

    def __init__(self):
        """Initialise a PubMLSTHandler object."""
        # Default values: Oxford (1) and Pasteur (2) schemes of A.baumannii species
        self.scheme_list = [
            self.base_url + "/pubmlst_abaumannii_seqdef/schemes/1",
            self.base_url + "/pubmlst_abaumannii_seqdef/schemes/2",
        ]
        self.scheme_paths = []
        self.scheme_mapping = {}

    def get_scheme_paths(self) -> dict:
        """
        Get the scheme paths in a dictionary.

        Returns:
            dict: A dictionary containing the scheme paths.
        """
        return scheme_list_to_dict(self.scheme_paths)

    def choose_schemes(self) -> None:
        """
        Changes the scheme list attribute to feature other schemes from another species.

        This function lets the user pick schemes to download all alleles that belong to it.
        The scheme has to be available in the database.
        """
        available_species = {}
        available_schemes = {}
        chosen_schemes = []
        counter = 1
        # retrieve all available species
        species_url = PubMLSTHandler.base_url
        for species_databases in requests.get(species_url, timeout=10).json():
            for database in species_databases["databases"]:
                if database["name"].endswith("seqdef"):
                    available_species[counter] = database["name"]
                    counter += 1
        # pick a species out of the available ones
        chosen_species = pick_species_number_from_db(available_species)

        counter = 1
        scheme_url = f"{species_url}/{chosen_species}/schemes"
        for scheme in requests.get(scheme_url, timeout=10).json()["schemes"]:
            # scheme["description"] stores the name of a scheme.
            # scheme["scheme"] stores the URL that is needed for downloading all loci.
            available_schemes[counter] = [scheme["description"], scheme["scheme"]]
            counter += 1

        # Selection process of available scheme from a species for download (doubles are caught!)
        while True:
            chosen_scheme = pick_scheme_number_from_db(available_schemes)
            if chosen_scheme not in chosen_schemes:
                chosen_schemes.append(chosen_scheme)
            choice = input(
                "Do you want to pick another scheme to download? (y/n):"
            ).lower()
            if choice != "y":
                break
        self.scheme_list = chosen_schemes

    def download_alleles(self, choice: False) -> None:
        """
        Downloads every allele FASTA-file from all loci of the scheme list attribute.

        This function sends API-GET requests to PubMLST.
        It downloads all alleles based on the scheme_list attribute.
        The default schemes are the Oxford and Pasteur schemes of A.baumannii

        Args:
            choice (bool): The decision to download different schemes, defaults to False.
        """
        if choice:  # pick an own scheme if not Oxford or Pasteur
            self.choose_schemes()  # changes the scheme_list attribute

        for scheme in self.scheme_list:
            scheme_json = requests.get(scheme, timeout=10).json()
            # We only want the name and the respective featured loci of a scheme
            scheme_name = scheme_json["description"]
            locus_list = scheme_json["loci"]

            species_name = scheme.split("_")[1]  # name = pubmlst_abaumannii_seqdef
            scheme_path = get_xspect_mlst_path() / species_name / scheme_name
            self.scheme_mapping[str(scheme_path)] = scheme
            self.scheme_paths.append(scheme_path)

            for locus_url in locus_list:
                # After using split the last part ([-1]) of the url is the locus name
                locus_name = locus_url.split("/")[-1]
                locus_path = (
                    get_xspect_mlst_path() / species_name / scheme_name / locus_name
                )

                if not locus_path.exists():
                    locus_path.mkdir(exist_ok=True, parents=True)

                alleles = requests.get(f"{locus_url}/alleles_fasta", timeout=10).text
                create_fasta_files(locus_path, alleles)

    def assign_strain_type_by_db(self) -> None:
        """
        Sends an API-POST-Request to the database for MLST without bloom filters.

        This function sends API-POST requests to PubMLST.
        It is a different way to determine strain types based on a BLAST-Search.
        This function is only used for testing and comparing results.
        """
        scheme_url = (
            str(pick_scheme(scheme_list_to_dict(self.scheme_list))) + "/sequence"
        )
        fasta_file = get_xspect_upload_path() / "Test.fna"
        with open(fasta_file, "r", encoding="utf-8") as file:
            data = file.read()
            payload = {  # Essential API-POST-Body
                "sequence": data,
                "filetype": "fasta",
            }
        response = requests.post(
            scheme_url, data=json.dumps(payload), timeout=10
        ).json()

        for locus, meta_data in response["exact_matches"].items():
            # meta_data is a list containing a dictionary, therefore [0] and then key value.
            # Example: 'Pas_fusA': [{'href': some URL, 'allele_id': '2'}]
            print(locus + ":" + meta_data[0]["allele_id"], end="; ")
        print("\nStrain Type:", response["fields"])

    def get_strain_type_name(self, highest_results: dict, post_url: str) -> str:
        """
        Send an API-POST request to PubMLST with the highest result of each locus as payload.

        This function formats the highest_result dict into an accepted input for the request.
        It gets a response from the site which is the strain type name.
        The name is based on the allele id with the highest score for each locus.
        Example of post_url for the oxford scheme of A.baumannii:
        https://rest.pubmlst.org/db/pubmlst_abaumannii_seqdef/schemes/1/designations

        Args:
            highest_results (dict): The allele ids with the highest kmer matches.
            post_url (str): The specific url for the scheme of a species

        Returns:
            str: The response (ST name or No ST found) of the POST request.
        """
        payload = {
            "designations": {
                locus: [{"allele": str(allele)}]
                for locus, allele in highest_results.items()
            }
        }

        response = requests.post(post_url + "/designations", json=payload, timeout=10)

        if response.status_code == 200:
            data = response.json()
            if "fields" in data:
                post_response = data["fields"]
                return post_response
            post_response = "No matching Strain Type found in the database. "
            post_response += "Possibly a novel Strain Type."
            return post_response
        post_response = "Error:" + str(response.status_code)
        post_response += response.text
        return post_response



from xspect.definitions import get_xspect_mlst_path
from xspect.mlst_feature.pub_mlst_handler import PubMLSTHandler

handler = PubMLSTHandler()


def test_download_default():
    """Tests the download functionality of alleles for the default case."""
    handler.download_alleles(False)
    species_path = get_xspect_mlst_path() / "abaumannii"
    first_scheme_path = species_path / "MLST (Oxford)"
    oxford_loci = [
        "Oxf_cpn60",
        "Oxf_gdhB",
        "Oxf_gltA",
        "Oxf_gpi",
        "Oxf_gyrB",
        "Oxf_recA",
        "Oxf_rpoD",
    ]
    assert sum(1 for _ in species_path.iterdir()) == 2  # 2 schemes
    for locus_path in sorted(first_scheme_path.iterdir()):
        locus_name = str(locus_path).split("/")[-1]
        assert locus_name in oxford_loci


def test_chosen_download(monkeypatch):
    """Tests the download functionality of alleles for a chosen scheme."""
    inputs = iter(["2", "1", "no"])  # Simulate input: [species, scheme, repeat]
    monkeypatch.setattr("builtins.input", lambda _: next(inputs))
    handler.download_alleles(True)
    assert handler.scheme_list == [
        "https://rest.pubmlst.org/db/pubmlst_abaumannii_seqdef/schemes/1"
    ]


def test_get_strain_type_name():
    """Tests the POST request that gets the strain type name based on the highest kmer results."""
    # url is from the oxford scheme of A.baumannii.
    post_url = "https://rest.pubmlst.org/db/pubmlst_abaumannii_seqdef/schemes/1"
    alleles = {  # Translates to ST 1 in the db.
        "Oxf_gyrB": 1,
        "Oxf_gltA": 1,
        "Oxf_gdhB": 1,
        "Oxf_recA": 1,
        "Oxf_gpi": 1,
        "Oxf_cpn60": 1,
        "Oxf_rpoD": 6,
    }

    strain_type_name = handler.get_strain_type_name(alleles, post_url)

    assert strain_type_name["ST"] == "1"



import pytest
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from xspect.mlst_feature.mlst_helper import pick_scheme, pick_scheme_from_models_dir
from xspect.mlst_feature.pub_mlst_handler import PubMLSTHandler
from xspect.models.probabilistic_filter_mlst_model import (
    ProbabilisticFilterMlstSchemeModel,
)
from xspect.definitions import get_xspect_model_path

handler = PubMLSTHandler()
PubMLSTHandler.download_alleles(handler, False)


@pytest.fixture
def filter_model():
    """Fixture for the ProbabilisticFilterMlstSchemeModel class."""
    return ProbabilisticFilterMlstSchemeModel(
        k_value=21,
        model_name="Test Filter",
        base_path=get_xspect_model_path() / "test",
        scheme_url="https://rest.pubmlst.org/db/pubmlst_abaumannii_seqdef/schemes/1",
    )


@pytest.fixture
def trained_filter_model(filter_model, monkeypatch):
    """Fixture for the ProbabilisticFilterModel class with trained model."""
    monkeypatch.setattr("builtins.input", lambda _: "1")
    scheme = pick_scheme(handler.get_scheme_paths())
    filter_model.fit(scheme)
    return filter_model


def test_model_initialization(
    filter_model,
):
    """Test the initialization of the ProbabilisticFilterMlstSchemeModel class."""
    assert filter_model.k_value == 21
    assert filter_model.model_name == "Test Filter"
    assert filter_model.model_type == "Strain"
    assert filter_model.fpr == 0.001


def test_model_save(trained_filter_model):
    """Test the save method of the ProbabilisticFilterMlstSchemeModel class."""
    trained_filter_model.save()
    assert (
        trained_filter_model.base_path / "MLST (Oxford)" / "MLST (Oxford).json"
    ).exists()


def test_fit(trained_filter_model):
    """Test the fit method of the ProbabilisticFilterMlstSchemeModel class."""
    assert len(trained_filter_model.indices) == 7  # Amount of cobs_structures
    expected_values = {
        "Oxf_cpn60": 265,
        "Oxf_gdhB": 381,
        "Oxf_gltA": 241,
        "Oxf_gpi": 541,
        "Oxf_gyrB": 352,
        "Oxf_recA": 254,
        "Oxf_rpoD": 286,
    }
    for locus, size in expected_values.items():
        # Important: size can be greater, because the database is updated regularly
        assert trained_filter_model.loci.get(locus) >= size


def test_predict(trained_filter_model, monkeypatch):
    """Test the predict method of the ProbabilisticFilterMlstSchemeModel class."""
    # Allele_ID_4 of Oxf_cpn60 with 401 kmers of length 21 each
    seq = Seq(
        "ATGAACCCAATGGATTTAAAACGCGGTATCGACATTGCAGTAAAAACTGTAGTTGAAAAT"
        "ATCCGTTCTATTGCTAAACCAGCTGATGATTTCAAAGCAATTGAACAAGTAGGTTCAATC"
        "TCTGCTAACTCTGATACTACTGTTGGTAAACTTATTGCTCAAGCAATGGAAAAAGTAGGT"
        "AAAGAAGGCGTAATCACTGTAGAAGAAGGTTCTGGCTTCGAAGACGCATTAGACGTTGTA"
        "GAAGGTATGCAGTTTGACCGTGGTTATATCTCTCCGTACTTTGCAAACAAACAAGATACT"
        "TTAACTGCTGAACTTGAAAATCCGTTCATTCTTCTTGTTGATAAAAAAATCAGCAACATT"
        "CGTGAATTGATTTCTGTTTTAGAAGCAGTTGCTAAAACTGGTAAACCACTTCTTATCATC"
        "G"
    )
    seq_record = SeqRecord(seq)
    monkeypatch.setattr("builtins.input", lambda _: "1")  # 1 = Oxford, 2 = Pasteur
    scheme = pick_scheme_from_models_dir()
    # [0] = Dict with the highest hits ([1] has all results which are not needed)
    res = trained_filter_model.predict(scheme, seq_record).hits.get("test")[0]
    allele_id = res.get("Strain type").get("Oxf_cpn60")

    assert allele_id.get("Allele_ID_4") == 401


def test_model_load(trained_filter_model):
    """Test the load method of the ProbabilisticFilterMlstSchemeModel class."""
    loaded_model = ProbabilisticFilterMlstSchemeModel.load(
        get_xspect_model_path() / "test" / "MLST" / "MLST (Oxford)"
    )
    assert loaded_model.k_value == 21
    assert loaded_model.model_name == "Test Filter"
    assert loaded_model.model_type == "Strain"
    assert len(loaded_model.indices) == 7
    expected_values = {
        "Oxf_cpn60": 265,
        "Oxf_gdhB": 381,
        "Oxf_gltA": 241,
        "Oxf_gpi": 541,
        "Oxf_gyrB": 352,
        "Oxf_recA": 254,
        "Oxf_rpoD": 286,
    }
    for locus, size in expected_values.items():
        # Important: size can be greater, because the database is updated regularly
        assert loaded_model.loci.get(locus) >= size


def test_sequence_splitter():
    """Test the sequence split method on an arbitrary short sequence."""
    model = ProbabilisticFilterMlstSchemeModel(
        k_value=4,
        model_name="Test Filter",
        base_path=get_xspect_model_path(),
        scheme_url="",
    )
    # len(seq) = 80; len(substring) = 20
    # k = 4 means each substring (except the first one) starts 3 (k - 1) base pairs earlier
    seq = "AGCTATTTCGCTGATGTCGACTGATCAAAAAGCCGGCGCGCTTTCGTATAGGCTAGCTACGACATACGATCGATCACTGA"
    res = model.sequence_splitter(seq, 20)
    # Does not assert 4 because of 3 additional base pairs when sliced (Last slice has 12 base pairs)
    assert len(res) == 5


def test_has_sufficient_score(trained_filter_model):
    """Tests if the kmer scores of a scheme are sufficient."""
    sufficient_dict = {
        "Scores": {
            "Oxf_cpn60": 265,
            "Oxf_gdhB": 381,
            "Oxf_gltA": 241,
            "Oxf_gpi": 541,
            "Oxf_gyrB": 352,
            "Oxf_recA": 254,
            "Oxf_rpoD": 286,
        }
    }

    not_sufficient_dict = {
        "Scores": {
            "Oxf_cpn60": 100,
            "Oxf_gdhB": 20,
            "Oxf_gltA": 6,
            "Oxf_gpi": 55,
            "Oxf_gyrB": 21,
            "Oxf_recA": 0,
            "Oxf_rpoD": 7,
        }
    }

    assert trained_filter_model.has_sufficient_score(
        sufficient_dict, trained_filter_model.avg_locus_bp_size
    )

    assert not trained_filter_model.has_sufficient_score(
        not_sufficient_dict, trained_filter_model.avg_locus_bp_size
    )
